/*
Deployment script for ExperienceForms DB rework
*/

GO
SET ANSI_NULLS, ANSI_PADDING, ANSI_WARNINGS, ARITHABORT, CONCAT_NULL_YIELDS_NULL, QUOTED_IDENTIFIER ON;

SET NUMERIC_ROUNDABORT OFF;

GO

SET XACT_ABORT ON
GO
SET TRANSACTION ISOLATION LEVEL READ COMMITTED
GO
BEGIN TRANSACTION


GO
IF NOT EXISTS
(
    SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = 'sitecore_forms_storage'
)
BEGIN
    PRINT N'Creating [sitecore_forms_storage] schema...';
    EXEC sp_executesql N'CREATE SCHEMA [sitecore_forms_storage]';
END


GO
IF NOT EXISTS
(
    SELECT schema_id FROM sys.types WHERE schema_id = SCHEMA_ID('sitecore_forms_storage') AND name = 'FormFieldData'
)
BEGIN
    PRINT N'Creating [sitecore_forms_storage].[FormFieldData] type...';
    CREATE TYPE [sitecore_forms_storage].[FormFieldData] AS TABLE
    (
        [Id] UNIQUEIDENTIFIER NOT NULL,
        [FieldDefinitionId] UNIQUEIDENTIFIER NOT NULL,
        [FieldName] NVARCHAR(256) NOT NULL,
        [Value] NVARCHAR(MAX) NULL,
        [ValueType] NVARCHAR(256) NULL
    );
END


GO
IF NOT EXISTS
(
    SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = 'sitecore_forms_storage' AND TABLE_NAME = 'FieldData'
)
BEGIN
    PRINT N'Creating [sitecore_forms_storage].[FieldData] table...';
    CREATE TABLE [sitecore_forms_storage].[FieldData]
    (
        [Id] UNIQUEIDENTIFIER NOT NULL,
        [FormEntryId] UNIQUEIDENTIFIER NOT NULL,
        [FieldDefinitionId] UNIQUEIDENTIFIER NOT NULL,
        [FieldName] NVARCHAR(256) NOT NULL,
        [Value] NVARCHAR(MAX) NULL,
        [ValueType] NVARCHAR(256) NULL,
        CONSTRAINT [PK_FieldData] PRIMARY KEY NONCLUSTERED ([Id] ASC)
    );
END


GO
IF NOT EXISTS
(
    SELECT object_id FROM sys.indexes WHERE object_id = OBJECT_ID('[sitecore_forms_storage].[FieldData]') AND name = 'IX_FieldData_FormEntryId'
)
BEGIN
    PRINT N'Creating [sitecore_forms_storage].[FieldData].[IX_FieldData_FormEntryId] non-clustered index...';
    CREATE NONCLUSTERED INDEX [IX_FieldData_FormEntryId]
        ON [sitecore_forms_storage].[FieldData]([FormEntryId]);
END


GO
IF NOT EXISTS
(
    SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = 'sitecore_forms_storage' AND TABLE_NAME = 'FormEntries'
)
BEGIN
    PRINT N'Creating [sitecore_forms_storage].[FormEntries] table...';
    CREATE TABLE [sitecore_forms_storage].[FormEntries]
    (
        [Id] UNIQUEIDENTIFIER NOT NULL,
        [FormDefinitionId] UNIQUEIDENTIFIER NOT NULL,
        [ContactId] UNIQUEIDENTIFIER NULL,
        [IsRedacted] BIT NOT NULL,
        [Created] DATETIME2(3) NOT NULL,
        CONSTRAINT [PK_FormEntry] PRIMARY KEY NONCLUSTERED ([Id] ASC)
    );
END


GO
DECLARE @ContactIdColumnExists BIT = CASE
    WHEN EXISTS
    (
        SELECT object_id FROM sys.columns WHERE object_id = OBJECT_ID(N'[sitecore_forms_storage].[FormEntries]') AND name = 'ContactId'
    )
    THEN 1
    ELSE 0
END
IF @ContactIdColumnExists = 0
    PRINT N'[ContactId] column for [sitecore_forms_storage].[FormEntries] table is missing...';

DECLARE @IsRedactedColumnExists BIT = CASE
    WHEN EXISTS
    (
        SELECT object_id FROM sys.columns WHERE object_id = OBJECT_ID(N'[sitecore_forms_storage].[FormEntries]') AND name = 'IsRedacted'
    )
    THEN 1
    ELSE 0
END
IF @IsRedactedColumnExists = 0
    PRINT N'[IsRedacted] column for [sitecore_forms_storage].[FormEntries] table is missing...';

IF @ContactIdColumnExists = 0 OR @IsRedactedColumnExists = 0
BEGIN
    PRINT N'Creating #FormEntries temp table...';
    CREATE TABLE #FormEntries
    (
        [Id] UNIQUEIDENTIFIER NOT NULL,
        [FormDefinitionId] UNIQUEIDENTIFIER NOT NULL,
        [ContactId] UNIQUEIDENTIFIER NULL,
        [IsRedacted] BIT NOT NULL,
        [Created] DATETIME2(3) NOT NULL
    );

    PRINT N'Inserting from [sitecore_forms_storage].[FormEntries] table to #FormEntries temp table...';
    INSERT INTO #FormEntries
    (
        [Id],
        [FormDefinitionId],
        [ContactId],
        [IsRedacted],
        [Created]
    )
    SELECT
        formEntriesInclNewColumnValues.[Id],
        formEntriesInclNewColumnValues.[FormDefinitionId],
        formEntriesInclNewColumnValues.[ContactId],
        formEntriesInclNewColumnValues.[IsRedacted],
        formEntriesInclNewColumnValues.[Created]
    FROM
    (
        SELECT
            NULL [ContactId],
            0 [IsRedacted]
    ) newColumnDefaultValues
    CROSS APPLY
    (
        SELECT
            [Id],
            [FormDefinitionId],
            [ContactId],
            [IsRedacted],
            [Created]
        FROM
            [sitecore_forms_storage].[FormEntries]
    ) formEntriesInclNewColumnValues;

    IF EXISTS
    (
        SELECT object_id FROM sys.foreign_keys WHERE parent_object_id = OBJECT_ID('[sitecore_forms_storage].[FieldData]') AND name = 'FK_FieldData_FormEntry'
    )
    BEGIN
        PRINT N'Dropping [sitecore_forms_storage].[FK_FieldData_FormEntry] foreign key constraint...';
        ALTER TABLE [sitecore_forms_storage].[FieldData] DROP CONSTRAINT [FK_FieldData_FormEntry];
    END

    PRINT N'Dropping [sitecore_forms_storage].[FormEntries] table...';
    DROP TABLE [sitecore_forms_storage].[FormEntries];

    PRINT N'Re-creating [sitecore_forms_storage].[FormEntries] table but with the previously missing columns...';
    CREATE TABLE [sitecore_forms_storage].[FormEntries]
    (
        [Id] UNIQUEIDENTIFIER NOT NULL,
        [FormDefinitionId] UNIQUEIDENTIFIER NOT NULL,
        [ContactId] UNIQUEIDENTIFIER NULL,
        [IsRedacted] BIT NOT NULL,
        [Created] DATETIME2(3) NOT NULL,
        CONSTRAINT [PK_FormEntry] PRIMARY KEY NONCLUSTERED ([Id] ASC)
    );

    PRINT N'Inserting from #FormEntries temp table to [sitecore_forms_storage].[FormEntries] table...';
    INSERT INTO [sitecore_forms_storage].[FormEntries]
    (
        [Id],
        [FormDefinitionId],
        [ContactId],
        [IsRedacted],
        [Created]
    )
    SELECT
        [Id],
        [FormDefinitionId],
        [ContactId],
        [IsRedacted],
        [Created]
    FROM
        #FormEntries

    PRINT N'Dropping #FormEntries temp table...';
    DROP TABLE #FormEntries;
END


GO
IF EXISTS
(
    SELECT object_id FROM sys.indexes WHERE object_id = OBJECT_ID('[sitecore_forms_storage].[FormEntries]') AND name = 'IX_FormEntry_Retrieval'
)
BEGIN
    PRINT N'Dropping [sitecore_forms_storage].[FormEntries].[IX_FormEntry_Retrieval] non-clustered index...';
    DROP INDEX [IX_FormEntry_Retrieval]
        ON [sitecore_forms_storage].[FormEntries]
END


GO
PRINT N'Creating [sitecore_forms_storage].[FormEntries].[IX_FormEntry_Retrieval] non-clustered index...';
CREATE NONCLUSTERED INDEX [IX_FormEntry_Retrieval]
    ON [sitecore_forms_storage].[FormEntries]([Id], [FormDefinitionId], [ContactId], [IsRedacted], [Created]);


GO
IF EXISTS
(
    SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = 'dbo' AND TABLE_NAME = 'FieldData'
)
BEGIN
    PRINT N'Inserting from [dbo].[FieldData] table to [sitecore_forms_storage].[FieldData] table...';
    INSERT INTO [sitecore_forms_storage].[FieldData]
    (
        [Id],
        [FormEntryId],
        [FieldDefinitionId],
        [FieldName],
        [Value],
        [ValueType]
    )
    SELECT
        [ID],
        [FormEntryID],
        [FieldItemID],
        [FieldName],
        [Value],
        [ValueType]
    FROM
        [dbo].[FieldData]
END


GO
IF EXISTS
(
    SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = 'dbo' AND TABLE_NAME = 'FormEntry'
)
BEGIN
    PRINT N'Inserting from [dbo].[FormEntry] table to [sitecore_forms_storage].[FormEntries] table...';
    INSERT INTO [sitecore_forms_storage].[FormEntries]
    (
        [Id],
        [FormDefinitionId],
        [ContactId],
        [IsRedacted],
        [Created]
    )
    SELECT
        [ID],
        [FormItemID],
        NULL,
        0,
        [Created]
    FROM
        [dbo].[FormEntry]
END


GO
IF NOT EXISTS
(
    SELECT object_id FROM sys.foreign_keys WHERE parent_object_id = OBJECT_ID('[sitecore_forms_storage].[FieldData]') AND name = 'FK_FieldData_FormEntry'
)
BEGIN
    PRINT N'Creating [sitecore_forms_storage].[FK_FieldData_FormEntry] foreign key constraint...';
    ALTER TABLE [sitecore_forms_storage].[FieldData] WITH NOCHECK
        ADD CONSTRAINT [FK_FieldData_FormEntry] FOREIGN KEY ([FormEntryId]) REFERENCES [sitecore_forms_storage].[FormEntries] ([Id]);
END


GO
IF EXISTS
(
    SELECT object_id FROM sys.indexes WHERE object_id = OBJECT_ID('[dbo].[FieldData]') AND name = 'FK_FieldData_FormEntry'
)
BEGIN
    PRINT N'Dropping [dbo].[FK_FieldData_FormEntry] foreign key constraint...';
    ALTER TABLE [dbo].[FieldData] DROP CONSTRAINT [FK_FieldData_FormEntry];
END


GO
IF EXISTS
(
    SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = 'dbo' AND TABLE_NAME = 'FieldData'
)
BEGIN
    PRINT N'Dropping [dbo].[FieldData] table...';
    DROP TABLE [dbo].[FieldData];
END


GO
IF EXISTS
(
    SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = 'dbo' AND TABLE_NAME = 'FormEntry'
)
BEGIN
    PRINT N'Dropping [dbo].[FormEntry] table...';
    DROP TABLE [dbo].[FormEntry];
END


GO
PRINT N'Updating [ValueType] column for [sitecore_forms_storage].[FieldData] table...';
GO
IF EXISTS
(
    SELECT [ValueType] FROM [sitecore_forms_storage].[FieldData] WHERE [ValueType] LIKE 'System.Collections.Generic.List%system.string%'
)
UPDATE
    [sitecore_forms_storage].[FieldData]
SET
    [ValueType] = 'System.Collections.Generic.List`1[System.String]'
WHERE
    [ValueType] LIKE 'System.Collections.Generic.List%system.string%'


GO
IF EXISTS
(
    SELECT object_id FROM sys.procedures WHERE object_id = OBJECT_ID('[sitecore_forms_storage].[FormData_Count]') AND type = 'P'
)
BEGIN
    PRINT N'Dropping [sitecore_forms_storage].[FormData_Count] stored procedure...';
    DROP PROCEDURE [sitecore_forms_storage].[FormData_Count]
END
GO
PRINT N'Creating [sitecore_forms_storage].[FormData_Count] stored procedure...';
GO
CREATE PROCEDURE [sitecore_forms_storage].[FormData_Count]
(
    @FormDefinitionId UNIQUEIDENTIFIER,
    @StartDate DATETIME2(3) = NULL,
    @EndDate DATETIME2(3) = NULL
)
WITH EXECUTE AS OWNER
AS
BEGIN
    SET NOCOUNT ON

    IF(@FormDefinitionId IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @FormDefinitionId cannot be null', 0
    END

    SELECT COUNT(*) AS [EntryCount]
    FROM [sitecore_forms_storage].[FormEntries]
    WHERE
        [FormDefinitionId] = @FormDefinitionId AND
        (@StartDate IS NULL OR [Created] >= @StartDate) AND
        (@EndDate IS NULL OR [Created] <= @EndDate)
END


GO
IF EXISTS
(
    SELECT object_id FROM sys.procedures WHERE object_id = OBJECT_ID('[sitecore_forms_storage].[FormData_Delete]') AND type = 'P'
)
BEGIN
    PRINT N'Dropping [sitecore_forms_storage].[FormData_Delete] stored procedure...';
    DROP PROCEDURE [sitecore_forms_storage].[FormData_Delete]
END
GO
PRINT N'Creating [sitecore_forms_storage].[FormData_Delete] stored procedure...';
GO
CREATE PROCEDURE [sitecore_forms_storage].[FormData_Delete]
(
    @FormDefinitionId UNIQUEIDENTIFIER,
    @StartDate DATETIME2(3) = NULL,
    @EndDate DATETIME2(3) = NULL
)
WITH EXECUTE AS OWNER
AS
BEGIN
    SET NOCOUNT ON

    IF(@FormDefinitionId IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @FormDefinitionId cannot be null', 0
    END

    DECLARE @Entries TABLE
    (
        [Id] UNIQUEIDENTIFIER NOT NULL
    )

    INSERT INTO
        @Entries
    SELECT
        [Id]
    FROM
        [sitecore_forms_storage].[FormEntries]
    WHERE
        [FormDefinitionId] = @FormDefinitionId AND
        (@StartDate IS NULL OR [Created] >= @StartDate) AND
        (@EndDate IS NULL OR [Created] <= @EndDate)

    DELETE
        [sitecore_forms_storage].[FieldData]
    WHERE
        [FormEntryId] IN (SELECT [Id] FROM @Entries)

    DELETE
        [sitecore_forms_storage].[FormEntries]
    WHERE
        [Id] IN (SELECT [Id] FROM @Entries)

    SELECT @@ROWCOUNT AS [DeletedEntryCount]
END


GO
IF EXISTS
(
    SELECT object_id FROM sys.procedures WHERE object_id = OBJECT_ID('[sitecore_forms_storage].[FormData_Redact]') AND type = 'P'
)
BEGIN
    PRINT N'Dropping [sitecore_forms_storage].[FormData_Redact] stored procedure...';
    DROP PROCEDURE [sitecore_forms_storage].[FormData_Redact]
END
GO
PRINT N'Creating [sitecore_forms_storage].[FormData_Redact] stored procedure...';
GO
CREATE PROCEDURE [sitecore_forms_storage].[FormData_Redact]
(
    @ContactId UNIQUEIDENTIFIER,
    @RedactedText VARCHAR(20)
)
WITH EXECUTE AS OWNER
AS
BEGIN
    SET NOCOUNT ON
    SET XACT_ABORT ON

    IF(@ContactId IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @ContactId cannot be null', 0
    END

    IF(@RedactedText IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @RedactedText cannot be null', 0
    END

    DECLARE
        @UpdatedEntryCount INT = 0

    BEGIN TRY
        BEGIN TRANSACTION

        DECLARE @RedactedFormEntryIds TABLE(Id UNIQUEIDENTIFIER)

        INSERT INTO
            @RedactedFormEntryIds
        SELECT
            [Id]
        FROM
            [sitecore_forms_storage].[FormEntries] 
        WHERE
            [ContactId] = @ContactId

        IF EXISTS
        (
            SELECT [Id] FROM @RedactedFormEntryIds
        )
        BEGIN
            UPDATE
                [sitecore_forms_storage].[FormEntries]
            SET
                [IsRedacted] = 1,
                [ContactId] = NULL
            WHERE
                [ContactId] = @ContactId
        END

        SELECT
            @UpdatedEntryCount = @@ROWCOUNT

        IF EXISTS
        (
            SELECT [Id] FROM [sitecore_forms_storage].[FieldData] WHERE [FormEntryId] IN (SELECT [Id] FROM @RedactedFormEntryIds)
        )
        BEGIN
            UPDATE
                [sitecore_forms_storage].[FieldData]
            SET
                [Value] = @RedactedText
            WHERE
                [FormEntryId] IN (SELECT [Id] FROM @RedactedFormEntryIds)
        END

        COMMIT TRANSACTION

        SELECT @UpdatedEntryCount AS [UpdatedEntryCount]

    END TRY

    BEGIN CATCH
        IF @@trancount > 0 ROLLBACK TRANSACTION
        ;THROW
    END CATCH
END


GO
IF EXISTS
(
    SELECT object_id FROM sys.procedures WHERE object_id = OBJECT_ID('[sitecore_forms_storage].[FormData_Retrieve]') AND type = 'P'
)
BEGIN
    PRINT N'Dropping [sitecore_forms_storage].[FormData_Retrieve] stored procedure...';
    DROP PROCEDURE [sitecore_forms_storage].[FormData_Retrieve]
END
GO
PRINT N'Creating [sitecore_forms_storage].[FormData_Retrieve] stored procedure...';
GO
CREATE PROCEDURE [sitecore_forms_storage].[FormData_Retrieve]
(
    @FormDefinitionId UNIQUEIDENTIFIER,
    @StartDate DATETIME2(3) = NULL,
    @EndDate DATETIME2(3) = NULL
)
WITH EXECUTE AS OWNER
AS
BEGIN
    SET NOCOUNT ON

    IF(@FormDefinitionId IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @FormDefinitionId cannot be null', 0
    END

    DECLARE @Entries TABLE
    (
        [Id] UNIQUEIDENTIFIER NOT NULL,
        [FormDefinitionId] UNIQUEIDENTIFIER NOT NULL,
        [Created] DATETIME2(3) NOT NULL
    )

    INSERT INTO
        @Entries
    SELECT
        [Id],
        [FormDefinitionId],
        [Created]
    FROM
        [sitecore_forms_storage].[FormEntries]
    WHERE
        [FormDefinitionId] = @FormDefinitionId AND
        (@StartDate IS NULL OR [Created] >= @StartDate) AND
        (@EndDate IS NULL OR [Created] <= @EndDate)

    SELECT
        [Id],
        [FormDefinitionId],
        [Created]
    FROM
        @Entries

    SELECT
        [Id],
        [FormEntryId],
        [FieldDefinitionId],
        [FieldName],
        [Value],
        [ValueType]
    FROM
        [sitecore_forms_storage].[FieldData]
    WHERE
        [FormEntryId] IN (SELECT [Id] FROM @Entries)
END


GO
IF EXISTS
(
    SELECT object_id FROM sys.procedures WHERE object_id = OBJECT_ID('[sitecore_forms_storage].[FormData_RetrieveContactUploadedFileIds]') AND type = 'P'
)
BEGIN
    PRINT N'Dropping [sitecore_forms_storage].[FormData_RetrieveContactUploadedFileIds] stored procedure...';
    DROP PROCEDURE [sitecore_forms_storage].[FormData_RetrieveContactUploadedFileIds]
END
GO
PRINT N'Creating [sitecore_forms_storage].[FormData_RetrieveContactUploadedFileIds] stored procedure...';
GO
CREATE PROCEDURE [sitecore_forms_storage].[FormData_RetrieveContactUploadedFileIds]
(
    @ContactId UNIQUEIDENTIFIER,
    @ValueType NVARCHAR(256)
)
WITH EXECUTE AS OWNER
AS
BEGIN
    SET NOCOUNT ON

    IF(@ContactId IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @ContactId cannot be null', 0
    END

    IF(@ValueType IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @ValueType cannot be null', 0
    END

    DECLARE @Entries TABLE
    (
        [Id] UNIQUEIDENTIFIER NOT NULL
    )

    INSERT INTO
        @Entries
    SELECT
        [Id]
    FROM
        [sitecore_forms_storage].[FormEntries]
    WHERE
        [ContactId] = @ContactId

    SELECT
        [Value]
    FROM
        [sitecore_forms_storage].[FieldData]
    WHERE
        [FormEntryId] IN (SELECT [Id] FROM @Entries) AND
        [ValueType] = @ValueType
END


GO
IF EXISTS
(
    SELECT object_id FROM sys.procedures WHERE object_id = OBJECT_ID('[sitecore_forms_storage].[FormData_RetrieveFieldValues]') AND type = 'P'
)
BEGIN
    PRINT N'Dropping [sitecore_forms_storage].[FormData_RetrieveFieldValues] stored procedure...';
    DROP PROCEDURE [sitecore_forms_storage].[FormData_RetrieveFieldValues]
END
GO
PRINT N'Creating [sitecore_forms_storage].[FormData_RetrieveFieldValues] stored procedure...';
GO
CREATE PROCEDURE [sitecore_forms_storage].[FormData_RetrieveFieldValues]
(
    @FormDefinitionId UNIQUEIDENTIFIER,
    @ValueType NVARCHAR(256),
    @StartDate DATETIME2(3) = NULL,
    @EndDate DATETIME2(3) = NULL
)
WITH EXECUTE AS OWNER
AS
BEGIN
    SET NOCOUNT ON

    IF(@FormDefinitionId IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @FormDefinitionId cannot be null', 0
    END

    IF(@ValueType IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @ValueType cannot be null', 0
    END

    DECLARE @Entries TABLE
    (
        [Id] UNIQUEIDENTIFIER NOT NULL
    )

    INSERT INTO
        @Entries
    SELECT
        [Id]
    FROM
        [sitecore_forms_storage].[FormEntries]
    WHERE
        [FormDefinitionId] = @FormDefinitionId AND
        (@StartDate IS NULL OR [Created] >= @StartDate) AND
        (@EndDate IS NULL OR [Created] <= @EndDate)

    SELECT
        [Value]
    FROM
        [sitecore_forms_storage].[FieldData]
    WHERE
        [FormEntryId] IN (SELECT [Id] FROM @Entries) AND
        [ValueType] = @ValueType
END


GO
IF EXISTS
(
    SELECT object_id FROM sys.procedures WHERE object_id = OBJECT_ID('[sitecore_forms_storage].[FormData_Store]') AND type = 'P'
)
BEGIN
    PRINT N'Dropping [sitecore_forms_storage].[FormData_Store] stored procedure...';
    DROP PROCEDURE [sitecore_forms_storage].[FormData_Store]
END
GO
PRINT N'Creating [sitecore_forms_storage].[FormData_Store] stored procedure...';
GO
CREATE PROCEDURE [sitecore_forms_storage].[FormData_Store]
(
    @FormEntryId UNIQUEIDENTIFIER,
    @FormDefinitionId UNIQUEIDENTIFIER,
    @ContactId UNIQUEIDENTIFIER,
    @IsRedacted BIT,
    @Created DATETIME2(3),
    @Fields [sitecore_forms_storage].[FormFieldData] READONLY
)
WITH EXECUTE AS OWNER
AS
BEGIN
    SET NOCOUNT ON
    SET XACT_ABORT ON

    IF(@FormEntryId IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @FormEntryId cannot be null', 0
    END

    IF(@FormDefinitionId IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @FormDefinitionId cannot be null', 0
    END

    IF(@IsRedacted IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @IsRedacted cannot be null', 0
    END

    IF(@Created IS NULL)
    BEGIN
        ;THROW 51000, N'Parameter @Created cannot be null', 0
    END

    DECLARE
        @UpdatedEntryCount INT = 0

    BEGIN TRY
        BEGIN TRANSACTION
        
        IF EXISTS
        (
            SELECT [Id] FROM [sitecore_forms_storage].[FormEntries] WHERE [Id] = @FormEntryId
        )
        BEGIN
            UPDATE
                [sitecore_forms_storage].[FormEntries]
            SET
                [FormDefinitionId] = @FormDefinitionId,
                [ContactId] = @ContactId,
                [IsRedacted] = @IsRedacted,
                [Created] = @Created
            WHERE
                [Id] = @FormEntryId
        END
        ELSE
        BEGIN
            INSERT INTO
                [sitecore_forms_storage].[FormEntries]
            (
                [Id],
                [FormDefinitionId],
                [ContactId],
                [IsRedacted],
                [Created]
            )
            VALUES
            (
                @FormEntryId,
                @FormDefinitionId,
                @ContactId,
                @IsRedacted,
                @Created
            )
        END

        SELECT
            @UpdatedEntryCount = @@ROWCOUNT

        MERGE
            [sitecore_forms_storage].[FieldData] WITH (TABLOCK, HOLDLOCK) AS [Target]
        USING
            @Fields AS [Source]
        ON
        (
            [Target].[Id] = [Source].[Id]
            OR
            (
                [Target].[FormEntryId] = @FormEntryId AND
                [Target].[FieldDefinitionId] = [Source].[FieldDefinitionId]
            )
        )
        WHEN MATCHED THEN
            UPDATE SET
                [FormEntryId] = @FormEntryId,
                [FieldDefinitionId] = [Source].[FieldDefinitionId],
                [FieldName] = [Source].[FieldName],
                [Value] = [Source].[Value],
                [ValueType] = [Source].[ValueType]
        WHEN NOT MATCHED THEN
            INSERT
            (
                [Id],
                [FormEntryId],
                [FieldDefinitionId],
                [FieldName],
                [Value],
                [ValueType]
            )
            VALUES
            (
                [Source].[Id],
                @FormEntryId,
                [Source].[FieldDefinitionId],
                [Source].[FieldName],
                [Source].[Value],
                [Source].[ValueType]
            )
        ;

        COMMIT TRANSACTION

        SELECT @UpdatedEntryCount AS [UpdatedEntryCount]

    END TRY
    BEGIN CATCH
        IF @@trancount > 0 ROLLBACK TRANSACTION
        ;THROW
    END CATCH
END


GO
IF EXISTS
(
    SELECT object_id FROM sys.procedures WHERE object_id = OBJECT_ID('[sitecore_forms_storage].[GrantLeastPrivilege]') AND type = 'P'
)
BEGIN
    PRINT N'Dropping [sitecore_forms_storage].[GrantLeastPrivilege] stored procedure...';
    DROP PROCEDURE [sitecore_forms_storage].[GrantLeastPrivilege]
END
GO
PRINT N'Creating [sitecore_forms_storage].[GrantLeastPrivilege] stored procedure...';
GO
CREATE PROCEDURE [sitecore_forms_storage].[GrantLeastPrivilege]
(
    @Name SYSNAME
)
WITH EXECUTE AS CALLER
AS
BEGIN
    SET NOCOUNT ON;

    IF (LEN(ISNULL(@Name, N'')) = 0)
    BEGIN
        ;THROW 51000, 'Parameter @Name cannot be null or empty', 0
    END

    IF (@Name LIKE N'%[^a-zA-Z0-9_]%')
    BEGIN
        ;THROW 51000, 'Parameter @Name cannot contain special characters', 0
    END

    EXECUTE ('
    -- Types
    grant execute on TYPE::[sitecore_forms_storage].[FormFieldData] TO ' + @Name + '

    -- Procs
    grant execute on [sitecore_forms_storage].[FormData_Count] TO ' + @Name + '
    grant execute on [sitecore_forms_storage].[FormData_Delete] TO ' + @Name + '
    grant execute on [sitecore_forms_storage].[FormData_Retrieve] TO ' + @Name + '
    grant execute on [sitecore_forms_storage].[FormData_RetrieveContactUploadedFileIds] TO ' + @Name + '
    grant execute on [sitecore_forms_storage].[FormData_RetrieveFieldValues] TO ' + @Name + '
    grant execute on [sitecore_forms_storage].[FormData_Store] TO ' + @Name + '
    grant execute on [sitecore_forms_storage].[FormData_Redact] TO ' + @Name);

END


GO

IF @@TRANCOUNT > 0
BEGIN
    PRINT N'The transacted portion of the database update succeeded.'
    COMMIT TRANSACTION
END
ELSE
    PRINT N'The transacted portion of the database update failed.'


GO